SULFUR_DEFAULT_VALENCE = 6
MIN_LONG_CYCLE_SIZE = 6

# normalization constants, statistics from 250k_rndm_zinc_drugs_clean.smi dataset
ZINC_LOGP_MEAN = 2.4570953396190123
ZINC_LOGP_STD = 1.434324401111988
ZINC_SA_MEAN = -3.0525811293166134
ZINC_SA_STD = 0.8335207024513095
ZINC_CYCLE_MEAN = -0.0485696876403053
ZINC_CYCLE_STD = 0.2860212110245455
